require "pgkingdom/columns_helper"

module PgKingdom
  # Class, which represents table in database. Contains information
  # about table schema, name, columns (with types and options), constraints,
  # primary and foreign keys. Also provides methods to generate {#create}({#create!}),
  # {#drop}({#drop!}) SQL statements.
  class Table
    # Returns schema name.
    # @return [String|Symbol]
    attr_reader :schema

    # Returns table name.
    # @return [String|Symbol]
    attr_reader :name

    # Defines new Table with a given block.
    # @raise [SyntaxError] when block was not given of block is empty.
    # @return [Postruby::Table] table instance.
    def initialize(name, &attr_block)
      if attr_block.nil?
        raise SyntaxError.new "Table columns should be given within a block."
      end

      if name.is_a? Array
        @schema = name.first
        @name = name.last
      else
        @schema = :public
        @name = name
      end

      @columns = []
      @columns.extend ColumnsHelper

      instance_eval &attr_block

      unless @columns.any?
        raise SyntaxError.new "Given block should not be empty."
      end
    end

    # Acts as DSL to define table columns.
    # @raise [RuntimeError] when column type is not supported or column was redeclared for current table.
    def method_missing(type, *opts)
      unless @columns.available? type
        raise "Unsupported column datatype #{type} for table #{@name}."
      end

      col_name = opts.shift
      if @columns.defined? col_name
        raise "Column #{col_name} with type #{type} has been redeclared for table #{@name}."
      end

      @columns.add col_name, type, opts
    end

    # Helper, which returns table path <schema>.<table_name>.
    # @return [String] path including schema name and table name.
    def table_path; "#{@schema}.#{@name}" end

    # Generates SQL to create table in the database.
    # When safe is true and table already exists it allows to
    # skip this action.
    # @return [String] SQL statement.
    def create(safe=true)
      allow_skip = safe ? " IF NOT EXISTS" : ""
      <<~SQL
      CREATE TABLE#{allow_skip} #{table_path} (
        #{@columns.join}
      );
      SQL
    end

    # Generates SQL to drop table in the database.
    # When safe is true and table doesn't exist it allows to skip this
    # action.
    # @return [String] SQL statement.
    def drop(safe=true)
      <<~SQL
      DROP TABLE#{safe ? " IF EXISTS" : ""} #{table_path};
      SQL
    end

    # Helper, which calls create with safe=false.
    # @return [String] SQL statement.
    def create!; create(false) end

    # Helper, which calls drop with safe=false.
    # @return [String] SQL statement.
    def drop!; drop(false) end

    # Helper, checks if column was defined in Table.
    # @return [Boolean]
    def column_defined?(name) @columns.defined? name end

    # Generates SQL query to check if column exists in table.
    # @return [String] SQL statement.
    def column_exists?(name, type=nil)
      # TODO: write pgSQL query
    end

    # Generates SQL to check if table exists in the database.
    # @return [String] SQL statement.
    def table_exists?
      "SELECT to_regclass('#{table_path}') IS NOT NULL AS table_exists;"
    end
  end
end
